/*
** Module   : CallMode.java
** Abstract : Specification for a parameter mode, when adding it via a CALL:SET-PARAMETER or
**            legacy ParameterList class.
**
** Copyright (c) 2019-2021, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description--------------------------------
** 001 CA  20190710 Extracted from Call.java.
** 002 GES 20210430 Moved APPEND into the ParameterOption.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * A container for the parameter's configured mode.
 */
public class CallMode
{
   /** Flag indicating INPUT mode. */
   public boolean input;
   
   /** Flag indicating OUTPUT mode. */
   public boolean output;
   
   /** Flag indicating INPUT-OUTPUT mode. */
   public boolean inputOutput;
   
   /** Parameter passing mode BY-VALUE, BIND or BY-REFERENCE option. */
   public ParameterOption mode = ParameterOption.NONE;
   
   /**
    * Given a string representation of the parameter's mode, interpret it and create a new
    * {@link CallMode} instance.
    * 
    * @param    mode
    *           The mode's string representation.
    *           
    * @return   {@code null} if the representation is not valid; the new instance, otherwise.
    */
   public static CallMode buildMode(character mode)
   {
      if (mode.isUnknown())
      {
         return null;
      }
      
      CallMode cmode = new CallMode();
      cmode.mode = ParameterOption.NONE;
      
      String smode = mode.getValue().toUpperCase();
      
      String append = "-APPEND";
      String byRef  = "-BY-REFERENCE"; // TODO: in OO, you can have 'input-output by-reference'
      String byVal  = "-BY-VALUE";
      String bind   = "-BIND";
      
      int optionLen = 0;
      
      if (smode.endsWith(append))
      {
         cmode.mode = ParameterOption.APPEND;
         optionLen = append.length();
      }
      else if (smode.endsWith(byRef))
      {
         cmode.mode = ParameterOption.BY_REFERENCE;
         optionLen = byRef.length();
      }
      else if (smode.endsWith(byVal))
      {
         cmode.mode = ParameterOption.BY_VALUE;
         optionLen = byVal.length();
      }
      else if (smode.endsWith(bind))
      {
         cmode.mode = ParameterOption.BIND;
         optionLen = bind.length();
      }
      
      if (optionLen != 0)
      {
         smode = smode.substring(0, smode.length() - optionLen);
      }
      
      switch (smode)
      {
         case "INPUT":
            cmode.input = true;
            break;
         case "OUTPUT":
            cmode.output = true;
            break;
         case "INPUT-OUTPUT":
            cmode.inputOutput = true;
            break;
         default:
            // invalid mode
            return null;
      }
      
      return cmode;
   }
   
   /**
    * Create a 1-char representation of this mode, as used by the {@link ControlFlowOps} APIs.
    * 
    * @return   "I" for INPUT, "O" for OUTPUT, "U" for INPUT-OUTPUT.
    */
   @Override
   public String toString()
   {
      return input ? "I" : output ? "O" : "U";
   }
}
